/*
 * browserwindow.cpp
 *
 * (c) Copyright Chumby Industries, 2010-2011
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QVBoxLayout>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#include <QWebSettings>
#include <QCleanlooksStyle>
#include <QWebFrame>

#include <QWebHistory>
#include <QWebHistoryItem>
#include <QSslError>
#include <QWebElement>
#include <QRegExp>

#include "browserwindow.h"
#include "webview.h"

bool qt_sendSpontaneousEvent(QObject *receiver, QEvent *event)
{
    return QCoreApplication::sendSpontaneousEvent(receiver, event);
}

#define ZOOMFIT
//#define FRAGMENT
#define FLICK
#define USERAGENT_MOBILE 0
#define USERAGENT_DESKTOP 1

#define SCREEN_WIDTH 860
#define SCREEN_HEIGHT 480

class BrowserWindowPrivate {
    public:
        QString formId;
        QString formName;
        QString formAction;
        QString inputType;
        QString inputId;
        QString inputName;
        QString inputValue;
        QString inputTitle;

        WebView* webView;
};

webPage::webPage()
{
    QWebPage::QWebPage();
    m_userAgent = USERAGENT_MOBILE;
}

//
// override user-agent
//
QString webPage::userAgentForUrl( const QUrl &url ) const
{
    if( m_userAgent == USERAGENT_MOBILE )
    {
        return QString( "Mozilla/5.0 (%Platform%; %Security%; %Subplatform%; %Locale%) AppleWebKit/%WebKitVersion% (KHTML, like Gecko) %AppVersion Mobile Safari/%WebKitVersion%" );
    }
    else
    {
        return QString( "Mozilla/5.0 (%Platform%; %Security%; %Subplatform%; %Locale%) AppleWebKit/%WebKitVersion% (KHTML, like Gecko) %AppVersion Safari/%WebKitVersion%" );
    }
}

void webPage::setUserAgent( int ua )
{
    m_userAgent = ua;
}

int webPage::getUserAgent( void )
{
    return m_userAgent;
}

BrowserWindow::BrowserWindow (QGraphicsWidget *parent)
    : QGraphicsWidget(parent), d(new BrowserWindowPrivate)
{

    thread = new fifoThread;
    thread->start();
        
    connect( thread, SIGNAL(seturl(QString)), this, SLOT(setURL(const QString&)) );
    connect( thread, SIGNAL(sethtml(QString, QString)), this, SLOT(setHTML(const QString&, const QString&)) );
    connect( thread, SIGNAL(setzoom(unsigned long, qreal)), this, SLOT(setZoom(unsigned long, qreal)) );
    connect( thread, SIGNAL(getzoom(unsigned long)), this, SLOT(getZoom(unsigned long)) );
    connect( thread, SIGNAL(evaluatejavascript(unsigned long, QString)), this, SLOT(evaluateJavascript(unsigned long, const QString&)) );    
    connect( thread, SIGNAL(setgeometry(int, int, int, int)), this, SLOT(setgeometry(int, int, int, int)) );

    connect( thread, SIGNAL( back( void ) ),    this, SLOT( back( void ) ) );
    connect( thread, SIGNAL( forward( void ) ), this, SLOT( forward( void ) ) );
    connect( thread, SIGNAL( reload( void ) ),  this, SLOT( reload( void ) ) );
    connect( thread, SIGNAL( stop( void ) ),    this, SLOT( stop( void ) ) );
    connect( thread, SIGNAL( quit( void ) ),    this, SLOT( quit( void ) ) );
    
    m_seqNum = 0;
    m_pipe = NULL;
    
    // Setup WebView    
    d->webView = new WebView;

    connect( thread, SIGNAL( setuseragent( int ) ), d->webView->page(), SLOT( setUserAgent( int ) ) );
    //connect( thread, SIGNAL( getuseragentmode( int ) ), d->webView->page(), SLOT( getUserAgentMode(int ) ) );

    // enable linkClicked signal
    d->webView->page()->setLinkDelegationPolicy( QWebPage::DelegateAllLinks );
    
    d->webView->setStyle( new QCleanlooksStyle() );

    connect( d->webView, SIGNAL( inputFormFocused(  const QString&, const QString&, 
                                                    const QString&, const QString&, 
                                                    const QString&, const QString&,
                                                    const QString&, const QString& )),
             this, SLOT( inputFormFocused(  const QString&, const QString&, 
                                            const QString&, const QString&, 
                                            const QString&, const QString&, 
                                            const QString&, const QString& )) );
    
    connect( d->webView, SIGNAL( inputFormLostFocus()), this, SLOT(inputFormLostFocus() ) );
    connect( d->webView, SIGNAL( gotJavascriptEval( const QString & ) ), this, SLOT(gotJavascriptEval( const QString & ) ) );
    
//    connect( d->webView, SIGNAL( loadStarted( ) ), this, SLOT( loadStarted( void ) ) );
    connect( d->webView, SIGNAL( loadProgress( int ) ), this, SLOT( loadProgress( int ) ) );    
    connect( d->webView, SIGNAL( loadFinished( bool ) ), this, SLOT( loadComplete( bool ) ) );
    //connect( d->webView, SIGNAL( statusBarMessage( const QString & ) ), this, SLOT( statusBarMessage( const QString & ) ) );  // removed because statusBarMessage is always empty
    
    connect( d->webView, SIGNAL( titleChanged( const QString & ) ), this, SLOT( titleChanged( const QString & ) ) );
    connect( d->webView, SIGNAL( linkClicked( const QUrl & ) ), this, SLOT( linkClicked( const QUrl & ) ) );
    connect( d->webView, SIGNAL( urlChanged( const QUrl & ) ), this, SLOT( urlChanged( const QUrl & ) ) );
    connect( d->webView->page()->networkAccessManager(), SIGNAL( sslErrors( QNetworkReply*, const QList<QSslError> & ) ), this, SLOT( ignoreSslErrors( QNetworkReply*, const QList<QSslError> & ) ) );

    QWebSettings::globalSettings()->setAttribute( QWebSettings::PluginsEnabled, true );
    //QWebSettings::globalSettings()->setAttribute( QWebSettings::TiledBackingStoreEnabled, true );
    //QWebSettings::globalSettings()->setAttribute( QWebSettings::FrameFlatteningEnabled, true );
    
    m_scene = new QGraphicsScene;
    m_view = new QGraphicsView( m_scene );
  
    m_view->setFrameShape( QFrame::NoFrame);
    m_view->setWindowFlags( Qt::FramelessWindowHint );

    d->webView->page()->mainFrame()->setScrollBarPolicy( Qt::Vertical,   Qt::ScrollBarAlwaysOff );
    d->webView->page()->mainFrame()->setScrollBarPolicy( Qt::Horizontal, Qt::ScrollBarAlwaysOff );

    d->webView->page()->setPreferredContentsSize(QSize( SCREEN_WIDTH, SCREEN_HEIGHT ));
    d->webView->setResizesToContents( true );

    m_scene->addItem( d->webView );
    
    m_view->resize( SCREEN_WIDTH, SCREEN_HEIGHT );
    m_view->setRenderHints(QPainter::Antialiasing | QPainter::SmoothPixmapTransform);

    const QBrush qb( Qt::SolidPattern ); 
    m_view->setBackgroundBrush( qb );
    m_view->show();
    
#ifdef FLICK
    m_flickCharm = new FlickCharm( this );

    m_flickCharm->activateOn( m_view );
    m_flickCharm->setWebView( d->webView );
    m_flickCharm->setWebFrame( d->webView->page()->mainFrame() );
    m_flickCharm->setGraphicsScene( m_scene );

    m_flickActive = true;
#endif

    m_urlChanged = true;
    m_firstLoadProgress = true;

    thread->setInputPanel( m_view->viewport() );

    m_view->viewport()->setFocusProxy( 0 );
    
    loadZoomBlackList();
}

void BrowserWindow::loadZoomBlackList( void )
{
    QFile* file = new QFile("/psp/browserZoomBlackList");

    if( !file->open( QIODevice::ReadOnly | QIODevice::Text ) )
    {
        return;
    }
    
    QXmlStreamReader xml( file );
    
    while( !xml.atEnd() && !xml.hasError() )
    {
        QXmlStreamReader::TokenType token = xml.readNext();
        
        if( token == QXmlStreamReader::StartDocument )
        {
			continue;
		}
		
        if( xml.name() == "zoomblacklist" )
        {
            continue;
        }

        if( token == QXmlStreamReader::StartElement)
        {
            if( xml.name() == "site" )
            {
                int subdomains = 0;
                if( xml.attributes().hasAttribute( "subdomains" ) )
                {
                    subdomains = atoi( xml.attributes().value( "subdomains" ).toString().toLatin1().data() );
                }
                m_zoomBlackListMap.insert( xml.attributes().value("url").toString(), subdomains );
            }
        }
    }
}

QKineticScroller* BrowserWindow::installKineticScroller( QWidget *w )
{
    // set a new kinetic scroller
    if (QAbstractScrollArea *area = qobject_cast<QAbstractScrollArea *>(w))
    {
        QScrollAreaKineticScroller *newScroller = new QScrollAreaKineticScroller();
        area->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
        area->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

        newScroller->setWidget(area);
        return newScroller;
    }
    else if (QWebView *web = qobject_cast<QWebView *>(w))
    {
        QWebViewKineticScroller *newScroller = new QWebViewKineticScroller();
        newScroller->setWidget(web);
        return newScroller;
    }
    else
    {
        return 0;
    }
}

void BrowserWindow::ignoreSslErrors( QNetworkReply* qnr, const QList<QSslError> & sslErrors )
{
    bool ignoreErrors = true;
    
    foreach( const QSslError &error, sslErrors )
    {
        char errorVal[3] = "";
        snprintf( errorVal, 3, "%d", error.error() );

        MapType data;
        data["action"] = "ssl";
        data["url"] = qnr->url().toString();
        data["value"] = errorVal;
        data["warning"] = qPrintable(error.errorString());
        sendActionEvent( &data );

	//if( error.error() != QSslError::HostNameMismatch )
        //{
            // found a legitimate SSL error, so we shouldn't ignore it
            ignoreErrors = false;
        //}
    }

    if( ignoreErrors )
    {
        qnr->ignoreSslErrors();
    }
}

void BrowserWindow::assurePipe( void )
{
    if( m_pipe == NULL )
    {
        const char* recvPipe = "/tmp/recv_browser_event";            
        m_pipe = fopen( (char*)recvPipe, "w" );
        setlinebuf( m_pipe );
    }
}

void BrowserWindow::saveScrollPosition( void )
{
    QPoint scrollPosition( m_view->horizontalScrollBar()->value(), m_view->verticalScrollBar()->value() );
    QVariant qv( scrollPosition );
    d->webView->page()->history()->currentItem().setUserData( qv );
}

void BrowserWindow::restoreScrollPosition( void )
{
    QPoint scrollPos = d->webView->page()->history()->currentItem().userData().toPoint();
    m_view->horizontalScrollBar()->setValue( scrollPos.x() );
    m_view->verticalScrollBar()->setValue( scrollPos.y() );
}

BrowserWindow::~BrowserWindow() {
    delete d;
}

//
// Public Slots
//

void BrowserWindow::load (const QString& url)
{
    d->webView->load(url);
}

// responsible for letting the fifo thread know which widget to
// send events to when focus changes
void BrowserWindow::focusChanged( QWidget* old, QWidget* now )
{
    if( now )
    {
        // don't disable the flick scroller if we're focused on the viewport
        if( now != m_view->viewport() )
        {
            m_web_scroller = installKineticScroller( now );
#ifdef FLICK            
            m_flickCharm->deactivateFrom( m_view );
#endif
        }
    }
    else
    {
        delete m_web_scroller;
#ifdef FLICK
        m_flickCharm->activateOn( m_view );
#endif
    }
}


//
// Private Slots
//

void BrowserWindow::back( void )
{
    saveScrollPosition();
    d->webView->back();
}

void BrowserWindow::forward( void )
{
    saveScrollPosition();
    d->webView->forward();
}

void BrowserWindow::reload( void )
{
    saveScrollPosition();
    d->webView->reload();
}

void BrowserWindow::stop( void )
{
    d->webView->stop();

    // reset the viewable area once the page load has been stopped
    m_view->setSceneRect( m_scene->itemsBoundingRect() ); 
}

void BrowserWindow::loadStarted( void )
{
    MapType data;
    
    data["action"] = "loadstarted";
    sendActionEvent( &data );    
}

void BrowserWindow::loadProgress( int progress )
{
    if( m_urlChanged )
    {
        if( m_firstLoadProgress )
        {
            loadStarted();
            m_firstLoadProgress = false;
        }
        
        MapType data;
            
        data["action"] = "loadprogress";
        data["progress"] = QString::number( progress );
        sendActionEvent( &data );  
    }
}

void BrowserWindow::scrollToFragment( const QString& fragment )
{
    if( ! fragment.isEmpty() )
    {
        QWebElement document = d->webView->page()->currentFrame()->documentElement();
        //fprintf( stderr, "VSCROLL: %s\n", document.findFirst( "[href=\"#" + fragment + "\"]").toInnerXml().toLatin1().data() );
        //fprintf( stderr, "VSCROLL: %d\n", document.findFirst( "[href=\"#" + fragment + "\"]" ).geometry().y() );
        int vscroll = document.findFirst( "[href=\"#" + fragment + "\"]" ).geometry().y();

        //fprintf( stderr, "vscroll: %d\n", vscroll );
        m_view->verticalScrollBar()->setValue( vscroll );
    }
}

void BrowserWindow::loadComplete( bool ok )
{
    MapType data;

    if( m_urlChanged )
    {

#ifdef ZOOMFIT
        bool shouldZoom = true;

        QUrl url = d->webView->url();
        
        // make sure we're not on an empty url or about:blank
        if( ( url.host() != QString( "about:blank" ) ) &&
            ( url.host() != QString( "" ) ) )
        {
            QMap<QString, int>::iterator it;
    
            for( it = m_zoomBlackListMap.begin(); it != m_zoomBlackListMap.end(); ++it )
            {
                QString host = it.key();
                int subdomains = it.value();

                if( !subdomains )
                {
                    // we're looking for an exact subdomain match
                    if( host == url.host() )
                    {
                        // we found a url in the blacklist, so we shouldn't auto zoom
                        shouldZoom = false;
                        break;
                    }
                }
                else
                {
                    // we're looking for any subdomains
                    QRegExp rx( "*." + host );
                    rx.setPatternSyntax( QRegExp::Wildcard );
                    
                    if( ( rx.exactMatch( url.host() ) ) || host == url.host() )
                    {
                        // we found a host that ended in the url in the blacklist, so we shouldn't auto zoom
                        shouldZoom = false;
                        break;
                    }
                }
            }

            d->webView->setZoomFactor( 1 );
    
            if( shouldZoom )
            {
            
                QRectF bounds = m_scene->itemsBoundingRect();
        
                double foo = ( ( SCREEN_WIDTH - 1 ) / bounds.width() ) * 10000;
                int bar = (int)foo;
                qreal zoom = (double)bar / 10000;
        
                //fprintf(stderr, "bounds.width(): %f\nzoom: %f\nfoo: %f\nbar: %d\n", bounds.width(), zoom, foo, bar );
    
                // force zoom level to match width of bounding rectangle    
                d->webView->setZoomFactor( zoom );
            }
        }
#endif // ZOOMFIT

        m_urlChanged = false;

        // reset the viewable area once the page has loaded
        m_view->setSceneRect( m_scene->itemsBoundingRect() );

#ifdef FRAGMENT
        if( d->webView->url().fragment().isEmpty() )
        {
            // restore scroll position if we're not loading a fragment
            restoreScrollPosition();
        }
        else
        {
            scrollToFragment( d->webView->url().fragment() );
        }
#else
        restoreScrollPosition();
#endif // FRAGMENT
    }
    
    data["action"] = "loadcomplete";
    data["url"]    = d->webView->url().toString();
    data["title"]  = d->webView->title();

    if( ok )
    {
        data["error"] = "0";
    }
    else
    {
        data["error"] = "1";
    }
        
    sendActionEvent( &data );  
}

void BrowserWindow::titleChanged( const QString& text )
{
    MapType data;
    
    data["action"]  = "titlechanged";
    data["title"] = text;
    sendActionEvent( &data );
}

void BrowserWindow::linkClicked( const QUrl& url )
{
    MapType data;
    
    data["action"]  = "linkclicked";
    data["link"] = url.toString();
    sendActionEvent( &data );
}

void BrowserWindow::urlChanged( const QUrl& url )
{
    MapType data;
    
    // make sure we're not switching to an empty url or about:blank
    if( strcmp( url.toString().toLatin1().data(), "about:blank" ) && strcmp( url.toString().toLatin1().data(), "" ) )
    {
        m_urlChanged = true;
        m_firstLoadProgress = true;
    }
    
    data["action"]  = "urlchanged";
    data["url"] = url.toString();
    sendActionEvent( &data );
}

void BrowserWindow::statusBarMessage( const QString& text )
{
    MapType data;

    data["action"]  = "statusbarmessagechanged";
    data["message"] = text;
    sendActionEvent( &data );
}

void BrowserWindow::gotJavascriptEval( const QString &param )
{
    MapType data;
    
    data["action"]  = "gotjavascripteval";
    data["result"]  = param;
    sendActionEvent( &data );    
}

void BrowserWindow::evaluateJavascript( unsigned long forSeq, const QString& script )
{
    MapType data;
    
    data["action"]  = "javascriptevaluated";
    data["for_seq"] = QString::number( forSeq );
    data["result"]  = d->webView->evalJavascript( script );                 
    sendActionEvent( &data );    
}

void BrowserWindow::setgeometry( int x, int y, int width, int height )
{
    this->setGeometry( x, y, width, height );
}

void BrowserWindow::setHTML( const QString& html, const QString& baseUrl )
{
    d->webView->setHtml( html, QUrl( baseUrl ) );
}

void BrowserWindow::setZoom(unsigned long forSeq, qreal zoom )
{
    d->webView->setZoomFactor( zoom );
    m_view->setSceneRect( m_scene->itemsBoundingRect() );
    this->getZoom( forSeq );
}

void BrowserWindow::getZoom( unsigned long forSeq )
{
    MapType data;
    
    data["action"]  = "zoom";
    data["value"]   = QString::number( d->webView->zoomFactor() );
    data["for_seq"] = QString::number( forSeq );
    sendActionEvent( &data );
}

void BrowserWindow::setURL( const QString& url )
{
    saveScrollPosition();

    d->webView->setHtml( "<html><body bgcolor=\"black\"></body></html>\n" );
    m_view->horizontalScrollBar()->setValue( 0 );
    m_view->verticalScrollBar()->setValue( 0 );

    // force zoom level to match width of bounding rectangle
    d->webView->setZoomFactor( 1 );
    m_view->setSceneRect( m_scene->itemsBoundingRect() );

    d->webView->load( url );
#ifdef FLICK
    m_flickCharm->setWebFrame( d->webView->page()->mainFrame() );
#endif

    d->webView->setResizesToContents( true );
}

void BrowserWindow::inputFormFocused (  const QString& formId,
                                        const QString& formName,
                                        const QString& formAction,
                                        const QString& inputType,
                                        const QString& inputId,                         
                                        const QString& inputName,
                                        const QString& inputValue,
                                        const QString& inputTitle )
{
    d->formId = formId;
    d->formName = formName;
    d->formAction = formAction;
    d->inputType = inputType;
    d->inputId = inputId;
    d->inputName = inputName;
    d->inputValue = inputValue;
    d->inputTitle = inputTitle;

    MapType data;
    
    data["formid"] = formId;
    data["formname"] = formName;
    data["formaction"] = formAction;
    data["inputtype"] = inputType;
    data["inputid"] = inputId;
    data["inputname"] = inputName;
    data["inputvalue"] = inputValue;
    data["inputtitle"] = inputTitle;
    data["action"] = "inputformfocus";
    
    sendActionEvent( &data ) ;
}

void BrowserWindow::inputFormLostFocus( void )
{
    // only dismiss the keyboard if the webView gets focus
    if( d->webView->hasFocus() )
    {
        MapType data;
        
        data["action"] = "inputformlostfocus";
        sendActionEvent( &data );
    }
}

void BrowserWindow::quit( void )
{
    unlink( "/tmp/send_browser_event" );
    unlink( "/tmp/recv_browser_event" );

    printf( "Exiting...\n" );
    exit( 1 );
}

void BrowserWindow::sendActionEvent( MapType* data )
{
    ++m_seqNum;
    assurePipe();

    QString xml = "";
    QXmlStreamWriter writer( &xml );

    writer.writeStartElement( "controlpanel" );
    writer.writeAttribute( "seq", QString::number( m_seqNum ) );
    
    MapType::const_iterator end = data->end();
    for( MapType::const_iterator it = data->begin(); it != end; ++it )
    {
        writer.writeAttribute( it->first, it->second );
    }
    
    writer.writeEndElement();
    fprintf( m_pipe, "%s\n", xml.toLatin1().data() );
    fprintf( stderr, "wrote: %s\n", xml.toLatin1().data() );
}
