/*
 * browserwindow.h
 *
 * (c) Copyright Chumby Industries, 2010-2011
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef _BROWSER_WINDOW_H_
#define _BROWSER_WINDOW_H_

#include <QGraphicsWidget>
#include <QScreenCursor>

#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QGraphicsWebView>

#include "fifothread.h"
#include "flickcharm.h"

#include "qwebviewkineticscroller.h"
#include "qscrollareakineticscroller.h"


class BrowserWindowPrivate;

class webPage : public QWebPage
{
    Q_OBJECT
    
	public:
	    webPage();
	    
    	QString userAgentForUrl(const QUrl &url ) const;

	public Q_SLOTS:
    	void setUserAgent( int ua );
    	int getUserAgent( void );

    private:
        int m_userAgent;
};

class BrowserWindow : public QGraphicsWidget {
	Q_OBJECT

	public:
		BrowserWindow (QGraphicsWidget *parent = 0);
		~BrowserWindow();
		
	public Q_SLOTS:
		void load (const QString& url);
		void focusChanged( QWidget* old, QWidget* now );

	private Q_SLOTS:
		void ignoreSslErrors( QNetworkReply* qnr, const QList<QSslError> & sslErrors );
		void inputFormFocused (	const QString& formId,
								const QString& formName,
								const QString& formAction,
								const QString& inputType,
								const QString& inputId,							
								const QString& inputName,
								const QString& inputValue,
								const QString& inputTitle );

        void inputFormLostFocus( void );
        void gotJavascriptEval( const QString& param );
        void loadStarted( void );
        void loadProgress( int progress );
        void loadComplete( bool ok );
        void statusBarMessage( const QString& text );
        void titleChanged( const QString& text );
        void linkClicked( const QUrl& url );
        void urlChanged( const QUrl& url );

        void back( void );
        void forward( void );
        void reload( void );
        void stop( void );

        void setURL( const QString& url );
        void setZoom( unsigned long forSeq, qreal zoom );
        void getZoom( unsigned long forSeq );
        void setHTML( const QString& html, const QString& baseUrl );
        void evaluateJavascript( unsigned long forSeq, const QString& script );
        void setgeometry( int x, int y, int width, int height );
        
        void quit( void );
        		
	private:

        void loadZoomBlackList( void );
        QKineticScroller* installKineticScroller( QWidget *w );

        typedef std::map <QString, QString> MapType;
        void sendActionEvent( MapType* data );
        void assurePipe( void );
        void saveScrollPosition( void );
        void restoreScrollPosition( void );
        void scrollToFragment( const QString& fragment );

        QKineticScroller* m_web_scroller;

        BrowserWindowPrivate* d;
        unsigned long m_seqNum;
        FILE* m_pipe;
        fifoThread* thread;
        FlickCharm* m_flickCharm;
        bool m_flickActive;
        bool m_urlChanged;
        bool m_firstLoadProgress;
        QMap<QString, int> m_zoomBlackListMap;

        QGraphicsWebView* m_webView;
        QGraphicsScene* m_scene;
        QGraphicsView* m_view;
		
};

#endif

