/*
 * fifothread.cpp
 *
 * (c) Copyright Chumby Industries, 2010-2011
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "fifothread.h"

void fifoThread::_processXmlEvent( char* cmd )
{
    QXmlStreamReader xml;

    xml.addData( cmd );
    
    while( !xml.atEnd() )
    {
        xml.readNext();
        if( xml.name() == "browser" )
        {
            if( xml.attributes().hasAttribute( "action" ) )
            {
               if( "setgeometry" == xml.attributes().value("action").toString() )
                {
                    if( xml.attributes().hasAttribute( "width" ) && xml.attributes().hasAttribute( "height" ) )
                    {
//                        fprintf( stderr, "width:%d\theight:%d\n", atoi( xml.attributes().value("width").toString().toLatin1().data() ),
//                                                                  atoi( xml.attributes().value("height").toString().toLatin1().data() ) );
                        emit setgeometry( atoi( xml.attributes().value("x").toString().toLatin1().data() ),
                                          atoi( xml.attributes().value("y").toString().toLatin1().data() ),
                                          atoi( xml.attributes().value("width").toString().toLatin1().data() ),
                                          atoi( xml.attributes().value("height").toString().toLatin1().data() ) );
                    }
                }
                else if( "seturl" == xml.attributes().value("action").toString() )
                {
                    if( xml.attributes().hasAttribute( "url" ) )
                    {
                        fprintf( stderr, "url: %s\n", xml.attributes().value("url").toString().toLatin1().data() );
                        emit seturl( xml.attributes().value("url").toString() );
                    }
                }
                else if( "sethtml" == xml.attributes().value("action").toString() )
                {
                    if( xml.attributes().hasAttribute( "html" ) && xml.attributes().hasAttribute( "baseurl" ))
                    {
                        fprintf( stderr, "html: %s\nbaseurl: %s\n", xml.attributes().value("html").toString().toLatin1().data(),
                                                           xml.attributes().value("baseurl").toString().toLatin1().data() );
                        emit sethtml( xml.attributes().value("html").toString(), xml.attributes().value("baseurl").toString() );
                    }
                }
                else if( "setzoom" == xml.attributes().value("action").toString() )
                {
                    if( xml.attributes().hasAttribute( "zoom" ) && xml.attributes().hasAttribute( "seq" ) )
                    {
                        fprintf( stderr, "seq: %ld\tzoom: %f\n", atol( xml.attributes().value("seq").toString().toLatin1().data() ),
                                                       atof( xml.attributes().value("zoom").toString().toLatin1().data() ) );
                        emit setzoom( atol( xml.attributes().value("seq").toString().toLatin1().data() ),
                                      atof( xml.attributes().value("zoom").toString().toLatin1().data() ) );
                    }
                }
                else if( "getzoom" == xml.attributes().value("action").toString() )
                {
                    if( xml.attributes().hasAttribute( "seq" ) )
                    {
                        fprintf( stderr, "getzoom\n" );
                        emit getzoom( atol( xml.attributes().value("seq").toString().toLatin1().data() ) );
                    }
                }
                else if( "evaluatejavascript" == xml.attributes().value("action").toString() )
                {
                    if( xml.attributes().hasAttribute( "script" ) && xml.attributes().hasAttribute( "seq" ) )
                    {
                        fprintf( stderr, "script: %s\n", xml.attributes().value("script").toString().toLatin1().data() );
                        emit evaluatejavascript( atol( xml.attributes().value("seq").toString().toLatin1().data() ),
                                                 xml.attributes().value("script").toString() );
                    }
                }
                else if( "setuseragent" == xml.attributes().value("action").toString() )
                {
                    if( xml.attributes().hasAttribute( "mode" ) )
                    {
                        if( atol( xml.attributes().value("mode").toString().toLatin1().data() ) == 0 )
                        {
                            emit setuseragent( 0 );
                        }
                        if( atol( xml.attributes().value("mode").toString().toLatin1().data() ) == 1 )
                        {
                            emit setuseragent( 1 );
                        }                        
                    }
                }                
                else if( "back" == xml.attributes().value("action").toString() )
                {
                    emit back();
                }
                else if( "forward" == xml.attributes().value("action").toString() )
                {
                    emit forward();
                }
                else if( "reload" == xml.attributes().value("action").toString() )
                {
                    emit reload();
                }
                else if( "stop" == xml.attributes().value("action").toString() )
                {
                    emit stop();
                }
                else if( "quit" == xml.attributes().value("action").toString() )
                {
                    emit quit();
                }
            }
        }
    }
    
    if( xml.error() && xml.error() != QXmlStreamReader::PrematureEndOfDocumentError )
    {
        qWarning() << "XML ERROR:" << xml.lineNumber() << ": " << xml.errorString();
    }
}

void fifoThread::_sendMouseEvent( char* cmd )
{
    int x = 0;
    int y = 0;
    char mouseState;
    QMouseEvent *mouseEvent = NULL;

    sscanf( cmd, "%c%d,%d", &mouseState, &x, &y );

    if( ( mouseState == 'M' ) && ( ! m_mouseDown ) )
    {
//        fprintf( stderr, "fifothread: MM without MD, throwing out event.\n" );
        return;
    }
    
    QPoint p;
    p.setX( x );
    p.setY( y );
//    fprintf( stderr, "before - x: %d, y: %d\n", p.x(), p.y() );
    p = m_pInputPanel->mapFromGlobal( p ); //convert mouse position to local for child widgets (drop-down lists, scrollbars, etc.)
//    fprintf( stderr, "after - x: %d, y: %d, s: %c\n", p.x(), p.y(), mouseState );

    //fprintf( stderr, "FROM FLASH - x: %d, y: %d, state: %c\n", x, y, mouseState );
    
    switch( mouseState )
    {
        case 'D':
            // mouse down event
            m_mouseDown = true;
            mouseEvent = new QMouseEvent(QEvent::MouseButtonPress, p, Qt::LeftButton, Qt::LeftButton, Qt::NoModifier);
            break;
        case 'U':
            // mouse up event
            mouseEvent = new QMouseEvent(QEvent::MouseButtonRelease, p, Qt::LeftButton, Qt::LeftButton, Qt::NoModifier);
            m_mouseDown = false;
            break;
        case 'M':
            // mouse move event
            mouseEvent = new QMouseEvent(QEvent::MouseMove, p, Qt::LeftButton, Qt::LeftButton, Qt::NoModifier); // enable kinetic scrolling
            //mouseEvent = new QMouseEvent(QEvent::MouseMove, p, Qt::NoButton, Qt::NoButton, Qt::NoModifier); // disable kinetic scrolling
            break;
        default:
            // unrecognized event
            fprintf(stderr, "unrecognized mouse event: %s\n", cmd );
            return;
            break;
    }

    //QPoint foo = QCursor::pos();
    //QCursor::setPos ( p );
    
//    m_pInputPanel->grabMouse();

    // have to use postEvent() instead of sendEvent() because it is threadsafe.
//    fprintf( stderr, "m_pInputPanel: %p\n", m_pInputPanel );
//    QApplication::postEvent( m_pInputPanel->focusWidget(), mouseEvent );
    QApplication::postEvent( m_pInputPanel, mouseEvent );

    QCoreApplication::processEvents();
}

void fifoThread::_sendKeyEvent( char* cmd )
{
    QKeyEvent *key = NULL;
    bool validKey  = false;
    
    int len = strlen( cmd );
    cmd[len - 1] = '\0'; // nuke newline
        
    if( strlen( cmd ) == 1 )
    {
        key = new QKeyEvent(QEvent::KeyPress, (int)cmd, Qt::NoModifier, cmd);
        validKey = true;
    }
    else if( strlen( cmd ) >= 2 )
    {
        if( strlen( cmd ) == 2 )
        {
            if( ! strncmp( cmd, "01", 2 ) )
            {
                key = new QKeyEvent(QEvent::KeyPress, Qt::Key_Home, Qt::NoModifier, NULL );
                validKey = true;
            }
            else if( ! strncmp( cmd, "03", 2 ) )
            {
                key = new QKeyEvent(QEvent::KeyPress, Qt::Key_End, Qt::NoModifier, NULL );
                validKey = true;
            }
            else if( ! strncmp( cmd, "08", 2 ) )
            {
                key = new QKeyEvent(QEvent::KeyPress, Qt::Key_Backspace, Qt::NoModifier, NULL );
                validKey = true;
            }
            else if( ! strncmp( cmd, "7F", 2 ) )
            {
                key = new QKeyEvent(QEvent::KeyPress, Qt::Key_Delete, Qt::NoModifier, NULL );
                validKey = true;
            }
            else if( !strncmp( cmd, "09", 2 ) )
            {
                key = new QKeyEvent(QEvent::KeyPress, Qt::Key_Tab, Qt::NoModifier, NULL );
                validKey = true;                
            }
            else if( !strncmp( cmd, "0D", 2 ) )
            {
                key = new QKeyEvent(QEvent::KeyPress, Qt::Key_Enter, Qt::NoModifier, NULL );
                validKey = true;                
            }
        }
        else
        {
            long unsigned result;
            char c;
    
            // shoutout to Duane for reducing hex conversion into a single line
            for( result = 0; ( c = *cmd++ ); result = ( result << 4 ) + (( c < 'A' ) ? ( c - '0' ) :( ( c < 'a') ? ( c - 'A' + 10 ) : ( c - 'a' + 10 ) ) ) ) ;

            fprintf( stderr, "result = %lx\n", result );
            key = new QKeyEvent(QEvent::KeyPress, (int)result, Qt::NoModifier, NULL );
            validKey = true;
        }
    }
    else if( strlen( cmd ) == 4 )
    {
        // two byte unicode
        validKey = false;
    }
    
    if( validKey )
    {
        m_pInputPanel->grabKeyboard();
        QApplication::postEvent( m_pInputPanel, key );
        QCoreApplication::processEvents();
    }
}

void fifoThread::setInputPanel( QWidget* pInputPanel )
{
    m_pInputPanel = pInputPanel;
    fprintf( stderr, "m_pInputPanel: %p\n", m_pInputPanel );
}

void fifoThread::run()
{
    char line [1024];
    const char* sendPipe = "/tmp/send_browser_event";
    const char* recvPipe = "/tmp/recv_browser_event";
    m_mouseDown = false;

    int ret = mkfifo( sendPipe, 0666 );
    if( ret != 0)
    {
        // fifo already exists, delete it and try again
        unlink(sendPipe);
        ret = mkfifo( sendPipe, 0666 );
        if(ret != 0)
        {
            fprintf( stderr, "error: unable to create pipe %s\n", (char*)sendPipe );
            exit( 1 );
        }
    }
    
    ret = mkfifo( recvPipe, 0666 );
    if( ret != 0)
    {
        // fifo already exists, delete it and try again
        unlink(recvPipe);
        ret = mkfifo( recvPipe, 0666 );
        if(ret != 0)
        {
            fprintf( stderr, "error: unable to create pipe %s\n", (char*)recvPipe );
            exit( 1 );
        }
    }
       
    while( 1 )
    {
        FILE* pipe = fopen( (char*)sendPipe, "r" );
    
        if (pipe == NULL)
        {
            fprintf( stderr, "Could not open input pipe %s\n", (char*)sendPipe );
            break;
        }
    
        setlinebuf( pipe );
    
        while( 1 )
        {
            if( fgets( line, sizeof( line ), pipe ) == NULL )
            {
                fclose( pipe );
                break;
            }
            else
            {
                char* cmd = NULL;
               // fprintf( stderr, "result on pipe %s: %s\n", (char*)sendPipe, line );

                cmd = strstr( line, ":" );
                
                switch ( line[0] )
                {
                    case 'M':
                        _sendMouseEvent( cmd + 1 );
                        break;
                    case 'K':
//                        fprintf( stderr, "keyboard command: %s\n", cmd + 1 );
                        _sendKeyEvent( cmd + 1 );
                        break;
                    case 'X':
                        fprintf( stderr, "xml command: %s\n", cmd + 1 );
                        _processXmlEvent( cmd + 1);
                        break;
                    default:
                        fprintf( stderr, "invalid command: %s\n", cmd + 1 );
                        break;
                }
            }
        }

        usleep( 1000 );
    }
}
