#include <QWebFrame>
#include <stdio.h>
#include <QVariant>
#include <QWebElement>
#include <QWebPage>
#include "webview.h"
#include "browserwindow.h"

WebView::WebView (QGraphicsWidget *parent) 
	: QGraphicsWebView(parent)
{

    webPage *page = new webPage();
    // force the webview to user our webPage object so we can override the user-agent
    this->setPage( (QWebPage*)page );
    
    page->setUserAgent( 0 ); // force mobile browser

	connect( page->mainFrame(), SIGNAL( javaScriptWindowObjectCleared() ), this, SLOT( populateJavaScriptWindowObject() ) );
	connect( this, SIGNAL( loadFinished( bool ) ), this, SLOT( onLoadFinished( bool ) ) );
	connect( page, SIGNAL( frameCreated( QWebFrame* ) ), this, SLOT( onFrameCreated( QWebFrame* ) ) );
}

WebView::~WebView()
{
}

QString WebView::evalJavascript( const QString& script )
{
    return page()->mainFrame()->evaluateJavaScript(script).toString();
}

void WebView::formFocusedAddValue (	const QString& formId,
									const QString& formName,
									const QString& formAction,
									const QString& inputType,
									const QString& inputId,
									const QString& inputName,
									const QString& value)
{
	QString scriptSource;
	
	if (!inputId.isEmpty()) {
		scriptSource = "document.getElementById('" + inputId + "').value += '" + value + "';";
	} else if (!formId.isEmpty()) {
		scriptSource = 
			"var inputList = document.getElementById('" + formId + "').getElementsByTagName('" + inputType + "');"
			"for (var i = 0; i < inputList.length; ++i) {"
				"if (inputList[i].name == '" + inputName + "') {"
					"inputList[i].value += '" + value + "';"
					"break;"
				"}"
			"}";
	} else if (!formAction.isEmpty()) {
		scriptSource = 
			"var formList = document.getElementsByTagName('form');"
			"for (var i = 0; i < formList.length; ++i) {"
				"if (formList[i].action == '" + formAction + "') {"
					"var inputList = formList[i].getElementsByTagName('" + inputType + "');"
					"for (var j = 0; j < inputList.length; ++j) {"
						"if (inputList[j].name == '" + inputName + "') {"
							"inputList[j].value += '" + value + "';"
							"break;"
						"}"
					"}"
					"break;"
				"}"
			"}";
	} else if (!formName.isEmpty()) {
		scriptSource = 
			"var formList = document.getElementsByTagName('form');"
			"for (var i = 0; i < formList.length; ++i) {"
				"if (formList[i].name == '" + formName + "') {"
					"var inputList = formList[i].getElementsByTagName('" + inputType + "');"
					"for (var j = 0; j < inputList.length; ++j) {"
						"if (inputList[j].name == '" + inputName + "') {"
							"inputList[j].value += '" + value + "';"
							"break;"
						"}"
					"}"
					"break;"
				"}"
			"}";
	} else {
		scriptSource = 
			"var inputList = document.getElementsByTagName('" + inputType + "');"
			"for (var i = 0; i < inputList.length; ++i) {"
				"if (inputList[i].name == '" + inputName + "') {"
					"inputList[i].value += '" + value + "';"
					"break;"
				"}"
			"}";
	}
	
	page()->mainFrame()->evaluateJavaScript(scriptSource);
}

void WebView::formFocused(	const QString& formId,
                            const QString& formName,
							const QString& formAction,
							const QString& inputType,
							const QString& inputId,
							const QString& inputName,
                            const QString& inputValue,
                            const QString& inputTitle )

{
	emit inputFormFocused(formId, formName, formAction, inputType, inputId, inputName, inputValue, inputTitle );
}

void WebView::formLostFocus (void) {
	emit inputFormLostFocus();
}


void WebView::getResult( const QString &param )
{
    emit gotJavascriptEval( param );
}

void WebView::onFrameCreated( QWebFrame* frame )
{
    fprintf( stderr, "WebView::onFrameCreated()" );
    frame->addToJavaScriptWindowObject("__qWebViewWidget", this);
    addFormsFocusEvent( frame );
}

void WebView::onLoadFinished( bool ok )
{
	if( ok )
	{
		addFormsFocusEvent( page()->currentFrame() );
        page()->currentFrame()->addToJavaScriptWindowObject("__qWebViewWidget", this);
	}
}

void WebView::populateJavaScriptWindowObject (void)
{
	page()->mainFrame()->addToJavaScriptWindowObject("__qWebViewWidget", this);
}


void WebView::addFormsFocusEvent (QWebFrame* frame)
{

	QString scriptSource =
        "function chumbrowser_check_iframes()"
        "{"
            "var frames = document.getElementsByTagName(\"iframe\");"
            "for (var i = 0; i < frames.length; i++)"
            "{"
                "var d = frames[i].contentDocument;"
                "chumbrowser_addFocusHandler( 'input', d );"
                "chumbrowser_addFocusHandler( 'textarea', d );"
                "chumbrowser_addFocusHandler( 'search', d );"
            "}"
        "}"

		"function chumbrowser_addFocusHandler( tagName, doc )"
		"{"
			"var inputList = doc.getElementsByTagName( tagName );"
			"for (var i = 0; i < inputList.length; ++i)"
			"{"
                "var focusString = 'formFocused(\"'+tagName+'\", '+i+')';"
				"if (tagName == 'input')"
				"{"
					"var inputType = inputList[i].type;"
                    "if (inputType != 'password' && inputType != 'text' && inputType != 'search')"
						"continue;"
				"}"
                "inputList[i].setAttribute('onclick', focusString );"
                "inputList[i].setAttribute('onblur', '__qWebViewWidget.formLostFocus()');"				
            "}"
		"}"
		
		"function formFocused( tagName, i )"
		"{"
            "var inputList = document.getElementsByTagName( tagName );"
            "var formAction = inputList[i].form.action;"
            "var formName = inputList[i].form.name;"
            "var formId = inputList[i].form.id;"
            "var inputName = inputList[i].name;"
            "var inputValue = inputList[i].value;"
            "var inputId = inputList[i].id;"
            "var inputTitle = inputList[i].title;"

            "if (tagName == 'input')"
            "{"
                "var inputType = inputList[i].type;"
                "if (inputType != 'password' && inputType != 'text' && inputType != 'search')"

                    "return;"
            "}"
            "__qWebViewWidget.formFocused( formId, formName, formAction, inputType, inputId, inputName, inputValue, inputTitle );"
        "}"

        "chumbrowser_addFocusHandler( 'input', document );"
        "chumbrowser_addFocusHandler( 'textarea', document );"
        "chumbrowser_addFocusHandler( 'search', document );"
        
    	"chumbrowser_check_iframes();";

    frame->evaluateJavaScript(scriptSource);
}
